#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include <limits.h>
#include <time.h>

#include "InOut.h"
#include "QAP.h"
#include "timer.h"
#include "utilities.h"
#include "ants.h"
#include "ls.h"
#include "parse.h"


double *best_in_try;
long int *best_found_at;
double   *time_best_found;
double   *time_total_run;

long int n_try;             /* try counter */
long int n_tours;           /* counter of number constructed tours */
long int iteration;         /* iteration counter */
long int restart_iteration; /* remember iteration when restart was done if any */
double   restart_time;      /* remember time when restart was done if any */
long int max_tries = 1;     /* maximum number of independent tries - not in use in this version*/
long int max_tours;         /* maximum number of tour constructions in one try */
long int seed;

long int egoinitial;        /* initial evaluations for EGO model */
long int egoevals;          /* evaluations for EGO */
long int egomax;            /* maximum number of data points used to build the EGO model */

double   lambda;            /* Parameter to determine branching factor */
double   branch_fac;        /* If branching factor < branch_fac => update trails */

double   max_time;          /* maximal allowed run time of a try  */
double   time_used;         /* time used until some given event */
double   time_passed;       /* time passed until some moment*/
double   optimal;           /* optimal solution or bound to find */


double   mean_ants;         /* average tour length */
double   stddev_ants;       /* stddev of tour lengths */
double   branching_factor;  /* average node branching factor when searching */
double   found_branching;   /* branching factor when best solution is found */

long int found_best;        /* iteration in which best solution is found */
long int restart_found_best;/* iteration in which restart-best solution is found */

/* ------------------------------------------------------------------------ */

FILE *report, *comp_report, *stat_report;

char name_buf[LINE_BUF_LEN];
int  opt;
long int quiet_flag;           /* --quiet was given in the command-line.  */


void write_report( void )
/*    
      FUNCTION: output some info about trial (best-so-far solution quality, time)
      INPUT:    none
      OUTPUT:   none
      COMMENTS: none
*/
{
    printf("best %f, iteration: %ld, time %.2f\n",best_so_far_ant->tour_length,iteration,elapsed_time( VIRTUALT ));
    if (comp_report) 
        fprintf(comp_report,
                "best %f\t iteration %ld\t tours %ld\t time %.3f\n",
                best_so_far_ant->tour_length,iteration,n_tours,time_used);
}


static void
fprintf_parameters (FILE *stream)
{
    fprintf(stream,"max_tours\t\t %ld\n", max_tours);
    fprintf(stream,"max_time\t\t %.2f\n", max_time);
    fprintf(stream,"seed\t\t %ld\n", seed);
    fprintf(stream,"optimum\t\t\t %f\n", optimal);
    fprintf(stream,"n_ants\t\t\t %ld\n", n_ants);
    fprintf(stream,"egoinitial\t\t\t %ld\n", egoinitial);
    fprintf(stream,"egoevals\t\t\t %ld\n", egoevals);
    fprintf(stream,"nn_ants\t\t\t %ld\n", nn_ants);
    fprintf(stream,"alpha\t\t\t %.2f\n", alpha);
    fprintf(stream,"beta\t\t\t %.2f\n", beta);
    fprintf(stream,"rho\t\t\t %.2f\n", rho);
    fprintf(stream,"q_0\t\t\t %.2f\n", q0);
    fprintf(stream,"elitist_ants\t\t %ld\n", elitist_ants);
    fprintf(stream,"ras_ranks\t\t %ld\n", ras_ranks);
    fprintf(stream,"ls_flag\t\t\t %d\n", ls_flag);
    fprintf(stream,"nn_ls\t\t\t %ld\n", nn_ls);
    fprintf(stream,"dlb_flag\t\t %ld\n", dlb_flag);
    fprintf(stream,"as_flag\t\t\t %ld\n", as_flag);
    fprintf(stream,"eas_flag\t\t %ld\n", eas_flag);
    fprintf(stream,"ras_flag\t\t %ld\n", ras_flag);
    fprintf(stream,"mmas_flag\t\t %ld\n", mmas_flag);
    fprintf(stream,"bwas_flag\t\t %ld\n", bwas_flag);
    fprintf(stream,"acs_flag\t\t %ld\n", acs_flag);

}

void print_default_parameters() 
/*    
      FUNCTION: output default parameter settings
      INPUT:    none
      OUTPUT:   none
      COMMENTS: none
*/
{
    fprintf(stderr,"\nDefault parameter settings are:\n\n");
    fprintf_parameters (stderr);
}


void
set_default_as_parameters (void)
{
    assert (as_flag);
    opt_n_ants         = -1;    /* number of ants (-1 means instance size) */
    nn_ants        = 20;    /* number of nearest neighbours in tour construction */
    alpha          = 1.0;
    opt_beta           = 2.0;
    opt_rho            = 0.5;
    opt_q0            = 0.0;
    ras_ranks      = 0;
    elitist_ants   = 0;
}

void
set_default_eas_parameters (void)
{
    assert (eas_flag);
    opt_n_ants         = -1;    /* number of ants (-1 means instance size) */
    nn_ants        = 20;    /* number of nearest neighbours in tour construction */
    alpha          = 1.0;
    opt_beta           = 2.0;
    opt_rho            = 0.5;
    opt_q0            = 0.0;
    ras_ranks      = 0;
    elitist_ants   = n_ants;
}

void
set_default_ras_parameters (void)
{
    assert (ras_flag);
    opt_n_ants         = -1;    /* number of ants (-1 means instance size) */
    nn_ants        = 20;    /* number of nearest neighbours in tour construction */
    alpha          = 1.0;
    opt_beta           = 2.0;
    opt_rho            = 0.1;
    opt_q0            = 0.0;
    ras_ranks      = 6;
    elitist_ants   = 0;
}


void
set_default_bwas_parameters (void)
{
    assert (bwas_flag);
    opt_n_ants         = -1;    /* number of ants (-1 means instance size) */
    nn_ants        = 20;    /* number of nearest neighbours in tour construction */
    alpha          = 1.0;
    opt_beta           = 2.0;
    opt_rho            = 0.1;
    opt_q0            = 0.0;
    ras_ranks      = 0;
    elitist_ants   = 0;
}

void
set_default_mmas_parameters (void)
{
    assert (mmas_flag);
    opt_n_ants         = -1;    /* number of ants (-1 means instance size) */
    nn_ants        = 20;    /* number of nearest neighbours in tour construction */
    alpha          = 1.0;
    opt_beta           = 2.0;
    opt_rho            = 0.02;
    opt_q0            = 0.0;
    ras_ranks      = 0;
    elitist_ants   = 0;
}

void
set_default_acs_parameters (void)
{
    assert (acs_flag);
    opt_n_ants         = 10;    /* number of ants (-1 means instance size) */
    nn_ants        = 20;    /* number of nearest neighbours in tour construction */
    alpha          = 1.0;
    opt_beta           = 2.0;
    opt_rho            = 0.1;
    opt_q0            = 0.9;
    ras_ranks      = 0;
    elitist_ants   = 0;
}

void set_default_ls_parameters(void)
{
    assert (ls_flag);
    dlb_flag       = FALSE; /* TRUE; */  /* apply don't look bits in local search */
    nn_ls          = 20;    /* use fixed radius search in the 20 nearest neighbours */

    opt_n_ants         = 25;    /* number of ants */
    nn_ants        = 20;    /* number of nearest neighbours in tour construction */
    alpha          = 1.0;
    opt_beta           = 2.0;
    opt_rho            = 0.5;
    opt_q0            = 0.0;

    if (mmas_flag) {
        opt_n_ants = 25;
        opt_rho = 0.2;
        opt_q0 = 0.00;
    } else if (acs_flag) {
        opt_n_ants = 10;
        opt_rho = 0.1;
        opt_q0 = 0.98;
    } else if (eas_flag) {
        elitist_ants = n_ants;
    }
}

void set_default_parameters(void) 
/*    
      FUNCTION: set default parameter settings
      INPUT:    none
      OUTPUT:   none
      COMMENTS: none
*/
{
    ls_flag        = (ls_types) 0;  /* LS_THREE_OPT_FIRST; */     /* don't run local search by default THOMAS */
    dlb_flag       = FALSE;  /* don't apply don't look bits in local search */
    nn_ls          = 20;     /* use fixed radius search in the 20 nearest neighbours */
    opt_n_ants     = 25; /* number of ants */
    nn_ants        = 20;     /* number of nearest neighbours in tour construction */
    alpha          = 1.0;
    opt_beta           = 2.0;
    opt_rho            = 0.5;
    opt_q0            = 0.0;
    max_tours      = 0;
    seed           = (long int) time(NULL);
    max_time       = 10.0;
    optimal        = 1;
    branch_fac     = 1.00001;
    u_gb           = LONG_MAX;
    as_flag        = FALSE;
    eas_flag       = FALSE;
    ras_flag       = FALSE;
    mmas_flag      = TRUE; 
    bwas_flag      = FALSE; 
    acs_flag       = FALSE;
    ras_ranks      = 0;
    elitist_ants   = 0;
    /*EGO*/
    egoinitial     = 10;
    egoevals       = 1000;
    egomax          = 100;
}



void population_statistics ( void )
/*    
      FUNCTION:       compute some population statistics like average tour length, 
                      standard deviations, average distance, branching-factor and 
		      output to a file gathering statistics
      INPUT:          none
      OUTPUT:         none
      (SIDE)EFFECTS:  none
*/
{
    long int j, k;
    long int *l;
    double   pop_mean, pop_stddev, avg_distance = 0.0;
    
    l = (long int *)malloc(n_ants * sizeof(long int));
    for( k = 0 ; k < n_ants ; k++ ) {
	l[k] = ant[k].tour_length;
    }
    
    pop_mean = mean( l, n_ants);
    pop_stddev = std_deviation( l, n_ants, pop_mean );
    branching_factor = node_branching(lambda);
    
    for ( k = 0 ; k < n_ants-1 ; k++ ) 
	for ( j = k+1 ; j < n_ants ; j++) {
	    avg_distance += (double)distance_between_ants( &ant[k], &ant[j]);
	}
    avg_distance /= ((double)n_ants * (double)(n_ants-1) / 2.); 
    
    if (stat_report) {
        fprintf(stat_report,"%ld\t%.1f\t%.5f\t%.7f\t%.5f\t%.1f\t%.1f\t%.5f\n",
                iteration, pop_mean, pop_stddev, pop_stddev / pop_mean,
                branching_factor, (branching_factor - 1.) * (double)n,
                avg_distance, avg_distance / (double)n);
    }
    free (l);
}



double node_branching(double l) 
/*    
      FUNCTION:       compute the average node lambda-branching factor 
      INPUT:          lambda value 
      OUTPUT:         average node branching factor 
      (SIDE)EFFECTS:  none
      COMMENTS:       see the ACO book for a definition of the average node 
                      lambda-branching factor 
*/
{
  long int  i, m;
  double    min, max, cutoff;
  double    avg;
  double    *num_branches;

  num_branches = (double *) calloc(n, sizeof(double));

  for ( m = 0 ; m < n ; m++ ) {
    /* determine max, min to calculate the cutoff value */
    min = pheromone[m][0];
    max = pheromone[m][0];
    for ( i = 1 ; i < n ; i++ ) {
      if ( pheromone[m][i] > max )
	max = pheromone[m][i];
      if ( pheromone[m][i] < min )
	min = pheromone[m][i];
    }
    cutoff = min + l * (max - min);
    
    for ( i = 0 ; i < n ; i++ ) {    
      if ( pheromone[m][i] > cutoff )
	num_branches[m] += 1.;
    }
  }
  avg = 0.;
  for ( m = 0 ; m < n ; m++ ) {
    avg += num_branches[m];
  }
  free ( num_branches );
  /* Norm branching factor to minimal value 1 */
  return ( avg / (double) (n)  );
}



void output_solution( void ) 
/*    
      FUNCTION:       output a solution together with node coordinates
      INPUT:          none
      OUTPUT:         none
      COMMENTS:       not used in the default implementation but may be useful anyway
*/
{

    long int i;
    if (stat_report) {
        for ( i = 0 ; i < n ; i++ ) {
	  fprintf(stat_report," %ld",best_so_far_ant->tour[i]);
        }
        fprintf(stat_report, "\n"); 
    }
}



void exit_try( long int ntry ) 
/*    
      FUNCTION:       save some statistical information on a trial once it finishes
      INPUT:          trial number
      OUTPUT:         none
      COMMENTS:       
*/
{
  checkTour( best_so_far_ant->tour );
/*    printTourFile( best_so_far_ant->tour ); */

  printf("\n Best Solution in try %ld is %f\n",ntry, best_so_far_ant->tour_length);
  if (report)
      fprintf(report, "Best: %f\t Iterations: %6ld\t B-Fac %.5f\t Time %.2f\t Tot.time %.2f\n",
              best_so_far_ant->tour_length, found_best, found_branching,
              time_used, elapsed_time( VIRTUALT ));
  printf(" Best Solution was found after %ld iterations\n", found_best);

  best_in_try[ntry] = best_so_far_ant->tour_length;
  best_found_at[ntry] = found_best;
  time_best_found[ntry] = time_used;
  time_total_run[ntry] = elapsed_time( VIRTUALT );
  printf("\ntry %ld, Best %f, found at iteration %ld, found at time %f\n",
         ntry, best_in_try[ntry], best_found_at[ntry], time_best_found[ntry]);

  if (comp_report) fprintf(comp_report,"end try %ld\n\n",ntry);
  if (stat_report) fprintf(stat_report,"end try %ld\n\n",ntry);
  TRACE (output_solution();)
  if (report) fflush(report); 
  if (comp_report) fflush(comp_report); 
  if (stat_report) fflush(stat_report); 

}



void exit_program( void ) 
/*    
      FUNCTION:       save some final statistical information on a trial once it finishes
      INPUT:          none
      OUTPUT:         none
      COMMENTS:       
*/
{
  double best_tour_length, worst_tour_length;
  double   t_avgbest, t_stdbest, t_avgtotal, t_stdtotal;
  double   avg_sol_quality = 0., avg_cyc_to_bst = 0., stddev_best, stddev_iterations;

  best_tour_length = best_of_vectorr( best_in_try ,max_tries );
  worst_tour_length = worst_of_vectorr( best_in_try , max_tries );

  avg_cyc_to_bst = mean( best_found_at , max_tries );
  stddev_iterations = std_deviation( best_found_at, max_tries, avg_cyc_to_bst );

  avg_sol_quality = meanr( best_in_try , max_tries );
  stddev_best = std_deviationr( best_in_try, max_tries, avg_sol_quality);

  t_avgbest = meanr( time_best_found, max_tries );
  printf(" t_avgbest = %f\n", t_avgbest );
  t_stdbest = std_deviationr( time_best_found, max_tries, t_avgbest);

  t_avgtotal = meanr( time_total_run, max_tries );
  printf(" t_avgtotal = %f\n", t_avgtotal );
  t_stdtotal = std_deviationr( time_total_run, max_tries, t_avgtotal);

  if (report) {
      fprintf(report,"\nAverage-Best: %.2f\t Average-Iterations: %.2f", avg_sol_quality, avg_cyc_to_bst);
      fprintf(report,"\nStddev-Best: %.2f \t Stddev Iterations: %.2f", stddev_best, stddev_iterations);
      fprintf(report,"\nBest try: %f\t\t Worst try: %f\n", best_tour_length, worst_tour_length);
      fprintf(report,"\nAvg.time-best: %.2f stddev.time-best: %.2f\n", t_avgbest, t_stdbest);  
      fprintf(report,"\nAvg.time-total: %.2f stddev.time-total: %.2f\n", t_avgtotal, t_stdtotal); 
      
      if ( optimal > 0 ) {
          fprintf(report,
                  " excess best = %f, excess average = %f, excess worst = %f\n",
                  (double)(best_tour_length - optimal) / (double)optimal,
                  (double)(avg_sol_quality - optimal) / (double)optimal,
                  (double)(worst_tour_length - optimal) / (double)optimal);
      }
  }

  if (comp_report)
      fprintf(comp_report,"end problem %s\n",instance.name);
}



void init_program( long int argc, char *argv[] ) 
/*    
      FUNCTION:       initialize the program, 
      INPUT:          program arguments, needed for parsing commandline
      OUTPUT:         none
      COMMENTS:       
*/
{

  FILE *tsp_file;
  char temp_buffer[LINE_BUF_LEN];
  char buf[LINE_BUF_LEN], *c;
  long int i;
  
  printf(PROG_ID_STR);
  set_default_parameters();
  setbuf(stdout,NULL);
  parse_commandline(argc, argv);
  
  best_in_try = (double *) calloc(max_tries, sizeof(double));
  best_found_at = (long int *) calloc(max_tries, sizeof(long int));
  time_best_found = (double *) calloc(max_tries, sizeof(double));
  time_total_run = (double *) calloc(max_tries, sizeof(double));
  
  if( (tsp_file = fopen(name_buf, "r")) == NULL) {
    fprintf(stderr, "error opening input file %s\n",name_buf);
    exit(1);	
  }
  
  TRACE ( printf("read problem data  ..\n\n"); )
  /* read instance data */
  read_problem_size( tsp_file );
  /* read_best_known_value( tsp_file ); */
  instance.distance = read_matrix( tsp_file, n);
  VERYVERBOSE ( print_matrix( instance.distance ); )  
  instance.flow = read_matrix( tsp_file, n );
  VERYVERBOSE ( print_matrix( instance.flow ); )  

  /* generate output file names and open them; I assume Lunix, Unix directory separators "/" */
  if ( ( c = strrchr(name_buf,'/') ) != NULL ) {
    c++;
    i = 0;
    while ((buf[i] = *c) != '\0') {
      i++;
      c++;
    }
    printf("%s\n",buf);
    strcpy(instance.name, buf);
  }


  TRACE ( printf("\n .. done\n\n"); )

  if (opt_n_ants < 0) opt_n_ants = n;
  /* default setting for elitist_ants is 0; if EAS is applied and
     option elitist_ants is not used, we set the default to
     elitist_ants = n */
  if (eas_flag && elitist_ants <= 0) elitist_ants = n;

  nn_ls = MIN(n - 1, nn_ls);

  /* Set maximum value from option setting. */
  max_n_ants = MAX(opt_n_ants, end_n_ants);

  /* Set runtime value from option setting. */
  n_ants = opt_n_ants;
  beta = opt_beta;
  rho = opt_rho;
  q0 = opt_q0;

  assert (max_n_ants < MAX_ANTS-1);
  assert (nn_ants < MAX_NEIGHBOURS);
  assert (nn_ants > 0);
  assert (nn_ls > 0);

  assert (n_ants < MAX_ANTS-1);
  assert (nn_ants < MAX_NEIGHBOURS);
  assert (nn_ants > 0);
  assert (nn_ls > 0);
  
  if (!quiet_flag) {
      sprintf(temp_buffer,"best.%s",instance.name);
      TRACE ( printf("%s\n",temp_buffer); )
          report = fopen(temp_buffer, "w");
      sprintf(temp_buffer,"cmp.%s",instance.name);
      TRACE ( printf("%s\n",temp_buffer); )
          comp_report = fopen(temp_buffer, "w");
      sprintf(temp_buffer,"stat.%s",instance.name);
      TRACE ( printf("%s\n",temp_buffer); )
          stat_report = fopen(temp_buffer, "w");
  } else {
      report = NULL;
      comp_report = NULL;
      stat_report = NULL;
  }
   
  write_params();
  if (comp_report)
      fprintf(comp_report,"begin problem %s\n",name_buf);
  printf("allocate ants' memory ..\n\n");
  allocate_ants();
  printf(" .. done\n");
}




void printDist(void) 
/*    
      FUNCTION:       print distance matrix 
      INPUT:          none
      OUTPUT:         none
*/
{
  long int i,j;

  printf("Distance Matrix:\n");
  for ( i = 0 ; i < n ; i++) {
    printf("From %ld:  ",i);
    for ( j = 0 ; j < n - 1 ; j++ ) {
      printf(" %ld", instance.distance[i][j]);
    }
    printf("\n");
  }
  printf("\n");
}


void printHeur(void) 
/*    
      FUNCTION:       print heuristic information 
      INPUT:          none
      OUTPUT:         none
*/
{
  long int i, j;

  printf("Heuristic information:\n");
  for ( i = 0 ; i < n ; i++) {
    printf("From %ld:  ",i);
    for ( j = 0 ; j < n - 1 ; j++ ) {
      printf(" %.3f ", HEURISTIC(i,j));
    }
    printf(" %.3f\n", HEURISTIC(i,j));
    printf("\n");
  }
  printf("\n");
}



void printTrail(void) 
/*    
      FUNCTION:       print pheromone trail values 
      INPUT:          none
      OUTPUT:         none
*/
{
  long int i,j;

  printf("pheromone Trail matrix, iteration: %ld\n\n",iteration);
  for ( i = 0 ; i < n ; i++) {
    printf("From %ld:  ",i);
    for ( j = 0 ; j < n ; j++ ) {
      printf(" %.10f ", pheromone[i][j]);
      if ( pheromone[i][j] > 1.0 )
	printf("XXXXX\n");
    }
    printf("\n");
  }
  printf("\n");
}



void printTotal(void) 
/*    
      FUNCTION:       print values of pheromone times heuristic information 
      INPUT:          none
      OUTPUT:         none
*/
{
  long int i, j;

  printf("combined pheromone and heuristic info\n\n");
  for (i=0; i < n; i++) {
    for (j = 0; j < n - 1 ; j++) {
      printf(" %.15f &", total[i][j]);
      if ( total[i][j] > 1.0 )
	printf("XXXXX\n");
    }
    printf(" %.15f\n", total[i][n-1]);
    if ( total[i][n-1] > 1.0 )
      printf("XXXXX\n");
  }
  printf("\n");
}



void printProbabilities(void) 
/*    
      FUNCTION:       prints the selection probabilities as encountered by an ant 
      INPUT:          none
      OUTPUT:         none
      COMMENTS:       this computation assumes that no choice has been made yet. 
*/
{
  long int i, j;
  double   *p;
  double   sum_prob;

  printf("Selection Probabilities, iteration: %ld\n",iteration);
  p = (double *)calloc( n, sizeof(double) );

  for (i=0; i < n; i++) {
    printf("From %ld:  ",i);
    sum_prob = 0.;
    for ( j = 0 ; j < n ; j++) {
      if ( i == j )
	p[j] = 0.;
      else
	p[j] = total[i][j];
      sum_prob += p[j];
    }
    for ( j = 0 ; j < n ; j++) {
      p[j] = p[j] / sum_prob;
    }
    for ( j = 0 ; j < n-1 ; j++) {
      printf(" %.5f ", p[j]);
    }
    printf(" %.5f\n", p[n-1]);
    if (!(j % 26)) {
      printf("\n");
    }
    printf("\n");
  }
  printf("\n");
  free ( p );
}



void printTour( long int *t ) 
/*    
      FUNCTION:       print the tour *t
      INPUT:          pointer to a tour
      OUTPUT:         none
*/
{
    long int   i;

    printf("\n");
    for( i = 0 ; i < n ; i++ ) {
	if (!i%25)
	    printf("\n");
	printf("%ld ", t[i]);
    }
    printf("\n");
    printf("Tour length = %ld\n\n",compute_tour_length( t ));
}



void printTourFile( long int *t ) 
/*    
      FUNCTION:       print the tour *t to cmp.tsplibfile
      INPUT:          pointer to a tour
      OUTPUT:         none
*/
{
    long int   i;
    if (!comp_report) return;

    fprintf(comp_report,"begin solution\n");
    for( i = 0 ; i < n ; i++ ) {
	fprintf(comp_report,"%ld ", t[i]);
    }
    fprintf(comp_report,"\n");
    fprintf(comp_report,"Tour length %ld\n",compute_tour_length( t ));
    fprintf(comp_report,"end solution\n");
}



void checkTour( long int *t ) 
/*    
      FUNCTION:       make a simple check whether tour *t can be feasible
      INPUT:          pointer to a tour
      OUTPUT:         none
*/
{
    long int i, sum=0;

    for( i = 0 ; i < n ; i++ ) {
	sum += t[i];
    }

    if ( sum != (n-1) * n / 2 ) {
	fprintf(stderr,"Next tour must be flawed !!\n");
	printTour( t );
	exit(1);
    }
}



void write_params( void ) 
/*    
      FUNCTION:       writes chosen parameter settings in standard output and in 
                      report files 
      INPUT:          none
      OUTPUT:         none
*/
{
    fprintf(stdout, "\nParameter-settings: \n\n");
    fprintf_parameters (stdout);
    fprintf(stdout, "\n");

  if (report) {
      fprintf(report,"\nParameter-settings: \n\n");
      fprintf_parameters (report);
      fprintf(report,"\n");
  }

  if (comp_report) {
      fprintf(comp_report,"%s",PROG_ID_STR);
      fprintf(comp_report,"\nParameter-settings: \n\n");
      fprintf_parameters (comp_report);
      fprintf(comp_report,"\n");
  }
}











