#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include <limits.h>
#include <time.h>

#include "InOut.h"
#include "QAP.h"
#include "ants.h"
#include "ls.h"
#include "utilities.h"
#include "timer.h"
#include "kriging.h"

ant_struct *ant;
ant_struct *best_so_far_ant;
ant_struct *restart_best_ant;

double   **pheromone;
double   **total;

double   *prob_of_selection;

long int n_ants;      /* number of ants */
long int max_n_ants;
long int nn_ants;     /* length of nearest neighbor lists for the ants'
			 solution construction */

double rho;           /* parameter for evaporation */
double alpha;         /* importance of trail */
double beta;          /* importance of heuristic evaluate */
double q0;           /* probability of best choice in tour construction */


long int as_flag;     /* ant system */
long int eas_flag;    /* elitist ant system */
long int ras_flag;    /* rank-based version of ant system */
long int mmas_flag;   /* MAX-MIN ant system */
long int bwas_flag;   /* best-worst ant system */
long int acs_flag;    /* ant colony system */

long int elitist_ants;    /* additional parameter for elitist
			     ant system, no. elitist ants */

long int ras_ranks;       /* additional parameter for rank-based version 
			     of ant system */

double   trail_max;       /* maximum pheromone trail in MMAS */
double   trail_min;       /* minimum pheromone trail in MMAS */
long int u_gb;            /* every u_gb iterations update with best-so-far ant */

double   trail_0;         /* initial pheromone level in ACS and BWAS */



void allocate_ants ( void )
/*    
      FUNCTION:       allocate the memory for the ant colony, the best-so-far and 
                      the iteration best ant
      INPUT:          none
      OUTPUT:         none
      (SIDE)EFFECTS:  allocation of memory for the ant colony and two ants that 
                      store intermediate tours

*/
{
    long int i;
  
    if((ant = (ant_struct *) malloc(sizeof(ant_struct) * max_n_ants +
		     sizeof(ant_struct *) * max_n_ants	 )) == NULL){
	printf("Out of memory, exit.");
	exit(1);
    }
    for ( i = 0 ; i < max_n_ants ; i++ ) {
	ant[i].tour        = (long int *) calloc(n+1, sizeof(long int));
	ant[i].visited     = (char *) calloc(n, sizeof(char));
    }

    if((best_so_far_ant = (ant_struct *) malloc(sizeof( ant_struct ) )) == NULL){
	printf("Out of memory, exit.");
	exit(1);
    }
    best_so_far_ant->tour        = (long int *) calloc(n+1, sizeof(long int));
    best_so_far_ant->visited     = (char *) calloc(n, sizeof(char));

    if((restart_best_ant = (ant_struct *) malloc(sizeof( ant_struct ) )) == NULL){
	printf("Out of memory, exit.");
	exit(1);
    }
    restart_best_ant->tour        = (long int *) calloc(n+1, sizeof(long int));
    restart_best_ant->visited     = (char *) calloc(n, sizeof(char));

    if ((prob_of_selection = (double *) malloc(sizeof(double) * (nn_ants + 1))) == NULL) {
	printf("Out of memory, exit.");
	exit(1);
    }
    /* Ensures that we do not run over the last element in the random wheel.  */
    prob_of_selection[nn_ants] = HUGE_VAL;
}


double find_best( void ) 
/*    
      FUNCTION:       find the best ant of the current iteration
      INPUT:          none
      OUTPUT:         index of struct containing the iteration best ant
      (SIDE)EFFECTS:  none
*/
{
    double  min;
    long int   k, k_min;

    min = ant[0].tour_length;
    k_min = 0;
    for( k = 1 ; k < n_ants ; k++ ) {
	if( ant[k].tour_length < min ) {
	    min = ant[k].tour_length;
	    k_min = k;
	}
    }
    return k_min;
}



double find_worst( void ) 
/*    
      FUNCTION:       find the worst ant of the current iteration
      INPUT:          none
      OUTPUT:         pointer to struct containing iteration best ant
      (SIDE)EFFECTS:  none
*/
{
    double   max;
    long int   k, k_max;

    max = ant[0].tour_length;
    k_max = 0;
    for( k = 1 ; k < n_ants ; k++ ) {
	if( ant[k].tour_length > max ) {
	    max = ant[k].tour_length;
	    k_max = k;
	}
    }
    return k_max;
}



/************************************************************
 ************************************************************
Procedures for pheromone manipulation 
 ************************************************************
 ************************************************************/



void init_pheromone_trails( double initial_trail )
/*    
      FUNCTION:      initialize pheromone trails
      INPUT:         initial value of pheromone trails "initial_trail"
      OUTPUT:        none
      (SIDE)EFFECTS: pheromone matrix is reinitialized
*/
{
    long int i, j;

    TRACE ( printf(" init trails with %.15f\n",initial_trail); );

    /* Initialize pheromone trails */
    for ( i = 0 ; i < n ; i++ ) {
	for ( j =0 ; j <= i ; j++ ) {
	    pheromone[i][j] = initial_trail;
	    pheromone[j][i] = initial_trail;
	    total[i][j] = initial_trail;
	    total[j][i] = initial_trail;
	}
    }
}



void evaporation( void )
/*    
      FUNCTION:      implements the pheromone trail evaporation
      INPUT:         none
      OUTPUT:        none
      (SIDE)EFFECTS: pheromones are reduced by factor rho
*/
{ 
    long int    i, j;

    TRACE ( printf("pheromone evaporation\n"); );

    for ( i = 0 ; i < n ; i++ ) {
	for ( j = 0 ; j < n ; j++ ) {
	    pheromone[i][j] = (1 - rho) * pheromone[i][j];
	    if (pheromone[i][j] < EPSILON)
	      pheromone[i][j] = EPSILON;
	}
    }
}



/* THOMAS: CHANGED */
void global_update_pheromone( ant_struct *a )
/*    
      FUNCTION:      reinforces edges used in ant k's solution
      INPUT:         pointer to ant that updates the pheromone trail 
      OUTPUT:        none
      (SIDE)EFFECTS: pheromones of arcs in ant k's tour are increased
*/
{  
    long int i;
    double   d_tau;

    TRACE ( printf("global pheromone update\n"); );

    d_tau = 1.0 / (double) a->tour_length;
    TRACE ( printf("deposit pheromone: quality: %ld amount pheromone %.15lf\n", a->tour_length,d_tau); )
    for ( i = 0 ; i < n ; i++ ) {
	pheromone[a->tour[i]][i] += d_tau;
    }
}



/* THOMAS: CHANGED */
void global_update_pheromone_weighted( ant_struct *a, long int weight )
/*    
      FUNCTION:      reinforces edges of the ant's tour with weight "weight"
      INPUT:         pointer to ant that updates pheromones and its weight  
      OUTPUT:        none
      (SIDE)EFFECTS: pheromones of arcs in the ant's tour are increased
*/
{  
    long int      i;
    double        d_tau;

    TRACE ( printf("global pheromone update weighted\n"); );

    d_tau = (double) weight / (double) a->tour_length;
    TRACE ( printf("deposit pheromone: quality: %ld amount pheromone %.15lf\n", a->tour_length,d_tau); )
    for ( i = 0 ; i < n ; i++ ) {
	pheromone[a->tour[i]][i] += d_tau;
    }       
}



/* THOMAS: CHANGED; COULD BE REMOVED BUT THEN MORE EFFORT TO REMOVE total AND RENAME IN CODE */
void compute_total_information( void )
/*    
      FUNCTION: calculates heuristic info times pheromone for each arc
      INPUT:    none  
      OUTPUT:   none
*/
{
    long int     i, j;

    TRACE ( printf("compute total information\n"); );

    for ( i = 0 ; i < n ; i++ ) {
	for ( j = 0 ; j < n ; j++ ) {
	  total[i][j] = pow(pheromone[i][j],alpha);  /* * pow(HEURISTIC(i,j),beta); THOMAS: NO HEURISTIC IN CODE */
	}
    }
}



/****************************************************************
 ****************************************************************
Procedures implementing solution construction and related things
 ****************************************************************
 ****************************************************************/



void ant_empty_memory( ant_struct *a ) 
/*    
      FUNCTION:       empty the ants's memory regarding visited cities
      INPUT:          ant identifier
      OUTPUT:         none
      (SIDE)EFFECTS:  vector of visited cities is reinitialized to FALSE
*/
{
    long int   i;

    for( i = 0 ; i < n ; i++ ) {
	a->visited[i]=FALSE;
    }
}



void place_ant( ant_struct *a , long int step )
/*    
      FUNCTION:      place an ant on a randomly chosen initial city
      INPUT:         pointer to ant and the number of construction steps 
      OUTPUT:        none
      (SIDE)EFFECT:  ant is put on the chosen city
*/
{
    long int     rnd;

    rnd = (long int) (ran01( &seed ) * (double) n); /* random number between 0 .. n-1 */
    a->tour[step] = rnd; 
    a->visited[rnd] = TRUE;
}


void choose_location(  ant_struct *a, long int item /* long int individual, long int item */)
/*    
      FUNCTION:      choose a location where to put the item 
      INPUT:         number indicating the ant and the item to be placed
      OUTPUT:        none
      (SIDE)EFFECTS: places the item on a location loc
*/
{ 
    long int  i;
    double     rnd, partial_sum = 0., sum_prob = 0.;
    double     *prob;
    long int  loc;

    if ( (q0 > 0.0) && (ran01( &seed ) < q0)  ) {
	/* with a probability q0 make the best possible choice
	   according to pheromone trails and heuristic information */
	/* we first check whether q0 > 0.0, to avoid the very common case
	   of q0 = 0.0 to have to compute a random number, which is
	   expensive computationally */
	choose_best_next(a, item);
	return;
    }

    prob = (double *) malloc ( n * sizeof(double));
  
    DEBUG( assert ( item >= 0 && item < n ); )

    for ( i = 0 ; i < n ; i++) {
	if ( a->visited[i] ) 
	    prob[i] = 0.0;
	else {
	    prob[i] = total[item][i];
	    sum_prob = sum_prob + prob[i];
	} 
    }

    /* now select to which location assign next an object */
    rnd = ran01(&seed);
    rnd *= sum_prob;
    i = 0;
    partial_sum = prob[i];

    while ((partial_sum <= rnd) && ( i < n-1 )) {
	i++;
	partial_sum += prob[i]; 
    }

    loc = i;
    DEBUG( assert ( loc >= 0 && loc < n ); )
    a->visited[loc] = TRUE; /* colony[individual].occupied[loc] = TRUE; */
    a->tour[loc] = item; /* colony[individual].s[loc] = item; */
    free(prob);
}



void choose_best_next( ant_struct *a, long int item )
/*    
      FUNCTION:      chooses for an ant as the next city the one with
                     maximal value of heuristic information times pheromone 
      INPUT:         pointer to ant and the construction step
      OUTPUT:        none 
      (SIDE)EFFECT:  ant moves to the chosen city
*/
{ 
    long int   i, loc=0;
    double   value_best;

    DEBUG( assert ( item > 0 && item < n ); );
    value_best = -1.;             /* values in total matrix are always >= 0.0 */    
    for ( i = 0 ; i < n ; i++ ) {
	if ( a->visited[i] ) 
	    ; /* city already visited, do nothing */
	else {
	    if ( total[item][i] > value_best ) {
		loc = i;
		value_best = total[item][i];
	    }
	} 
    }
    DEBUG( assert ( 0 <= loc && loc < n); );
    DEBUG( assert ( value_best > 0.0 ); )
    DEBUG( assert ( a->visited[loc] == FALSE ); )
    a->tour[loc] = item;
    a->visited[loc] = TRUE;
}




void choose_closest_next( ant_struct *a, long int phase )
/*    
      FUNCTION:      Chooses for an ant the closest city as the next one 
      INPUT:         pointer to ant and the construction step "phase" 
      OUTPUT:        none 
      (SIDE)EFFECT:  ant moves to the chosen city
*/
{ 
    long int city, current_city, next_city, min_distance;
  
    next_city = n;
    DEBUG( assert ( phase > 0 && phase < n ); );
    current_city = a->tour[phase-1];
    min_distance = LONG_MAX;             /* Search shortest edge */    
    for ( city = 0 ; city < n ; city++ ) {
	if ( a->visited[city] ) 
	    ; /* city already visited */
	else {
	    if ( instance.distance[current_city][city] < min_distance) {
		next_city = city;
		min_distance = instance.distance[current_city][city];
	    }
	} 
    }
    DEBUG( assert ( 0 <= next_city && next_city < n); );
    a->tour[phase] = next_city;
    a->visited[next_city] = TRUE;
}






/****************************************************************
 ****************************************************************
Procedures specific to MAX-MIN Ant System
 ****************************************************************
****************************************************************/


/* THOMAS: CHANGED AVOID SYMMETRIC PHEROMONE MATRIX */
void check_pheromone_trail_limits( void )
/*    
      FUNCTION:      only for MMAS without local search: 
                     keeps pheromone trails inside trail limits
      INPUT:         none
      OUTPUT:        none
      (SIDE)EFFECTS: pheromones are forced to interval [trail_min,trail_max]
*/
{ 
    long int    i, j;

    TRACE ( printf("mmas specific: check pheromone trail limits\n"); );

    for ( i = 0 ; i < n ; i++ ) {
	for ( j = 0 ; j < n ; j++ ) {
	    if ( pheromone[i][j] < trail_min ) {
		pheromone[i][j] = trail_min;
	    } else if ( pheromone[i][j] > trail_max ) {
		pheromone[i][j] = trail_max;
	    }
	}
    }
}



/****************************************************************
 ****************************************************************
Procedures specific to Ant Colony System
 ****************************************************************
****************************************************************/



void global_acs_pheromone_update( ant_struct *a )
/*    
      FUNCTION:      reinforces the edges used in ant's solution as in ACS
      INPUT:         pointer to ant that updates the pheromone trail 
      OUTPUT:        none
      (SIDE)EFFECTS: pheromones of arcs in ant k's tour are increased
*/
{  
    long int i;
    double   d_tau;

    TRACE ( printf("acs specific: global pheromone update\n"); );

    d_tau = 1.0 / (double) a->tour_length;

    for ( i = 0 ; i < n ; i++ ) {

	pheromone[a->tour[i]][i] = (1. - rho) * pheromone[a->tour[i]][i] + rho * d_tau;

	total[a->tour[i]][i] = pow(pheromone[a->tour[i]][i], alpha);  /* * pow(HEURISTIC(a->tour[i],i),beta); */
    }
}



void local_acs_pheromone_update( ant_struct *a, long int i )
/*    
      FUNCTION:      removes some pheromone on edge just passed by the ant
      INPUT:         pointer to ant and number of constr. phase
      OUTPUT:        none
      (SIDE)EFFECTS: pheromones of arcs in ant k's tour are increased
      COMMENTS:      I did not do experiments with with different values of the parameter 
                     xi for the local pheromone update; therefore, here xi is fixed to 0.1 
		     as suggested by Gambardella and Dorigo for the TSP. If you wish to run 
		     experiments with that parameter it may be reasonable to use it as a 
		     commandline parameter
*/
{  
    
    /* still additional parameter has to be introduced */
    pheromone[a->tour[i]][i] = (1. - 0.1) * pheromone[a->tour[i]][i] + 0.1 * trail_0;

    total[a->tour[i]][i] = pow(pheromone[a->tour[i]][i], alpha); /* * pow(HEURISTIC([a->tour[i],i),beta); */

}



/****************************************************************
 ****************************************************************
Procedures specific to Best-Worst Ant System
 ****************************************************************
****************************************************************/



void bwas_worst_ant_update( ant_struct *a1, ant_struct *a2)
/*    
      FUNCTION:      uses additional evaporation on the arcs of iteration worst
                     ant that are not shared with the global best ant
      INPUT:         pointer to the worst (a1) and the best (a2) ant
      OUTPUT:        none
      (SIDE)EFFECTS: pheromones on some arcs undergo additional evaporation
*/
{  
    long int    i, j, h, pos, pred;
    long int    *pos2;        /* positions of cities in tour of ant a2 */ 

    TRACE ( printf("bwas specific: best-worst pheromone update\n"); );

    pos2 = (long int *) malloc(n * sizeof(long int));
    for ( i = 0 ; i < n ; i++ ) {
	pos2[a2->tour[i]] = i;
    }
 
    for ( i = 0 ; i < n ; i++ ) {
	j = a1->tour[i];
	h = a1->tour[i+1];
	pos = pos2[j];
	if (pos - 1 < 0)
	    pred = n - 1;
	else 
	    pred = pos - 1;
	if (a2->tour[pos+1] == h)
	    ; /* do nothing, edge is common with a2 (best solution found so far) */
	else if (a2->tour[pred] == h)
	    ; /* do nothing, edge is common with a2 (best solution found so far) */
	else {   /* edge (j,h) does not occur in ant a2 */       
	    pheromone[j][h] = (1 - rho) * pheromone[j][h];
	    pheromone[h][j] = (1 - rho) * pheromone[h][j];
	}
    }
    free ( pos2 );
}



void bwas_pheromone_mutation( void )
/*    
      FUNCTION: implements the pheromone mutation in Best-Worst Ant System
      INPUT:    none  
      OUTPUT:   none
*/
{
    long int     i, j, k;
    long int     num_mutations;
    double       avg_trail = 0.0, mutation_strength = 0.0, mutation_rate = 0.3;

    TRACE ( printf("bwas specific: pheromone mutation\n"); );

    /* compute average pheromone trail on edges of global best solution */
    for ( i = 0 ; i < n ; i++ ) {
	avg_trail +=  pheromone[best_so_far_ant->tour[i]][best_so_far_ant->tour[i+1]];
    }
    avg_trail /= (double) n;
  
    /* determine mutation strength of pheromone matrix */ 
    /* FIXME: we add a small value to the denominator to avoid any
       potential division by zero. This may not be fully correct
       according to the original BWAS. */
    if ( max_time > 0.1 )
	mutation_strength = 4. * avg_trail * (elapsed_time(VIRTUALT) - restart_time) / (max_time - restart_time + 0.0001);
    else if ( max_tours > 100 )
	mutation_strength = 4. * avg_trail * (iteration - restart_iteration) 
            / (max_tours - restart_iteration + 1);
    else
	printf("apparently no termination condition applied!!\n");

    /* finally use fast version of matrix mutation */
    mutation_rate = mutation_rate / n * nn_ants;
    num_mutations = n * mutation_rate / 2;   
    /* / 2 because of adjustment for symmetry of pheromone trails */
 
    if ( restart_iteration < 2 )
	num_mutations = 0; 

    for ( i = 0 ; i < num_mutations ; i++ ) {
	j =   (long int) (ran01( &seed ) * (double) n);
	k =   (long int) (ran01( &seed ) * (double) n);
	if ( ran01( &seed ) < 0.5 ) {
	    pheromone[j][k] += mutation_strength;
	    pheromone[k][j] = pheromone[j][k];
	}
	else {
	    pheromone[j][k] -= mutation_strength;
	    if ( pheromone[j][k] <= 0.0 ) {
		pheromone[j][k] = EPSILON;
	    }
	    pheromone[k][j] = pheromone[j][k]; 
	}
    }
}



/**************************************************************************
 **************************************************************************
Procedures specific to the ant's tour manipulation other than construction
***************************************************************************
 **************************************************************************/



void copy_from_to(ant_struct *a1, ant_struct *a2) 
{
/*    
      FUNCTION:       copy solution from ant a1 into ant a2
      INPUT:          pointers to the two ants a1 and a2 
      OUTPUT:         none
      (SIDE)EFFECTS:  a2 is copy of a1
*/
    int   i;
  
    a2->tour_length = a1->tour_length;
    for ( i = 0 ; i < n ; i++ ) {
	a2->tour[i] = a1->tour[i];
    }
    a2->tour[n] = a2->tour[0];
}



/* THOMAS: CHANGED, GENERATE RANDOM SOLUTION AND GIVE BACK OBJECTIVE */
double nn_tour() 
/*    
      FUNCTION:       generate some nearest neighbor tour and compute tour length
      INPUT:          none
      OUTPUT:         none
      (SIDE)EFFECTS:  needs ant colony and one statistic ants
*/
 
{
    long int i;
    double help;
    long int *objects;

    ant_empty_memory( &ant[0] );
    objects = generate_random_permutation( n );

    for ( i = 0 ; i < n ; i++ ) {
      ant[0].tour[i] = objects[i];
    }
    ant[0].tour_length = surrogate_tour_length(ant[0].tour, "ei");
    help = ant[0].tour_length;

    ant_empty_memory( &ant[0] );
    free (objects);

    return help;
}



long int distance_between_ants( ant_struct *a1, ant_struct *a2)
/*    
      FUNCTION: compute the distance between the tours of ant a1 and a2
      INPUT:    pointers to the two ants a1 and a2
      OUTPUT:   distance between ant a1 and a2
*/
{  
    long int    i, j, h, pos, pred;
    long int    distance;
    long int    *pos2;        /* positions of cities in tour of ant a2 */ 

    pos2 = (long int *)malloc(n * sizeof(long int));
    for ( i = 0 ; i < n ; i++ ) {
	pos2[a2->tour[i]] = i;
    }

    distance = 0;
    for ( i = 0 ; i < n ; i++ ) {
	j = a1->tour[i];
	h = a1->tour[i+1];
	pos = pos2[j];
	if (pos - 1 < 0)
	    pred = n - 1;
	else 
	    pred = pos - 1;
	if (a2->tour[pos+1] == h)
	    ; /* do nothing, edge is common with best solution found so far */
	else if (a2->tour[pred] == h)
	    ; /* do nothing, edge is common with best solution found so far */
	else {   /* edge (j,h) does not occur in ant a2 */       
	    distance++;
	}
    }
    free ( pos2 );
    return distance;
}


